/**
 * @file ngx_http_waf_module_check.h
 * @brief 检查诸如 IP，URL 等是否命中规则。
*/


#ifndef NGX_HTTP_WAF_MODLULE_CHECK_H
#define NGX_HTTP_WAF_MODLULE_CHECK_H


#include <ngx_config.h>
#include <ngx_core.h>
#include <ngx_http.h>
#include <ngx_regex.h>
#include <ngx_inet.h>
#include <uthash.h>
#include <math.h>
#include <ngx_http_waf_module_macro.h>
#include <ngx_http_waf_module_type.h>
#include <ngx_http_waf_module_util.h>
#include <ngx_http_waf_module_ip_trie.h>
#include <ngx_http_waf_module_lru_cache.h>
#include <ngx_http_waf_module_captcha.h>


/**
 * @brief 用来挂载到清理请求资源的函数，主要用来存储和获取 ngx_http_waf_ctx_t。
*/
void ngx_http_waf_handler_cleanup(void *data);

/**
 * @defgroup check 规则匹配模块
 * @brief 检查诸如 IP，URL 等是否命中规则。
 * @addtogroup check 规则匹配模块
 * @{
*/


/**
 * @brief 检查客户端 IP 地址是否在白名单中。
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED IP 地址在白名单中。
 * @retval NOT_MATCHED IP 地址不在白名单中。
*/
ngx_int_t ngx_http_waf_handler_check_white_ip(ngx_http_request_t* r);


/**
 * @brief 检查客户端 IP 地址是否在黑名单中。
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED IP 地址在黑名单中。
 * @retval NOT_MATCHED IP 地址不在黑名单中。
*/
ngx_int_t ngx_http_waf_handler_check_black_ip(ngx_http_request_t* r);


/**
 * @brief 检查客户端 IP 地址的访问频次（60 秒内）是否超出了限制。
 * @return 如果超出 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 超出限制。
 * @retval NOT_MATCHED 未超出限制。
*/
ngx_int_t ngx_http_waf_handler_check_cc(ngx_http_request_t* r);


/**
 * @brief 检查 URL 是否在白名单中。
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在白名单中。
 * @retval NOT_MATCHED 不在白名单中
*/
ngx_int_t ngx_http_waf_handler_check_white_url(ngx_http_request_t* r);


/**
 * @brief 检查 URL 是否在黑名单中
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在黑名单中。
 * @retval NOT_MATCHED 不在黑名单中
*/
ngx_int_t ngx_http_waf_handler_check_black_url(ngx_http_request_t* r);


/**
 * @brief 检查请求参数是否在黑名单中
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在黑名单中。
 * @retval NOT_MATCHED 不在黑名单中
*/
ngx_int_t ngx_http_waf_handler_check_black_args(ngx_http_request_t* r);


/**
 * @brief 检查 UserAgent 是否在黑名单中
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在黑名单中。
 * @retval NOT_MATCHED 不在黑名单中
*/
ngx_int_t ngx_http_waf_handler_check_black_user_agent(ngx_http_request_t* r);


/**
 * @brief 检查 Referer 是否在白名单中
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在白名单中。
 * @retval NOT_MATCHED 不在白黑名单中
*/
ngx_int_t ngx_http_waf_handler_check_white_referer(ngx_http_request_t* r);


/**
 * @brief 检查 Referer 是否在黑名单中
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在黑名单中。
 * @retval NOT_MATCHED 不在黑名单中
*/
ngx_int_t ngx_http_waf_handler_check_black_referer(ngx_http_request_t* r);


/**
 * @brief 检查 Cookie 是否在黑名单中
 * @return 如果在返回 MATCHED，反之返回 NOT_MATCHED。
 * @retval MATCHED 在黑名单中。
 * @retval NOT_MATCHED 不在黑名单中
*/
ngx_int_t ngx_http_waf_handler_check_black_cookie(ngx_http_request_t* r);


/**
 * @brief 检查请求体内容是否存在于黑名单中，存在则拦截，反之放行。
*/
ngx_int_t ngx_http_waf_handler_check_black_post(ngx_http_request_t* r);


/**
 * @brief 测试数组内的所有正则
 * @param[in] str 被测试的字符串
 * @param[in] array 包含若干个正则的数组
 * @param[in] rule_type 触发规则时的规则类型
 * @param[in] cache 检测时所使用的缓存管理器
 * @return 如果匹配到返回 NGX_HTTP_WAF_MATCHED，反之则为 NGX_HTTP_WAF_NOT_MATCHED。
*/
ngx_int_t ngx_http_waf_regex_exec_arrray(ngx_http_request_t* r, 
                                         ngx_str_t* str, 
                                         ngx_array_t* array, 
                                         const u_char* rule_type, 
                                         lru_cache_t* cache);

/**
 * @}
*/


#endif
